# ruff: noqa: UP006, UP035
#
import collections.abc
import re
import sys
from collections import defaultdict
from collections.abc import Iterable, Mapping, Sequence
from contextlib import contextmanager
from typing import (
    TYPE_CHECKING,
    AbstractSet,
    Annotated,
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Literal,
    Optional,
    Set,
    TypedDict,
    TypeVar,
    Union,
)

import dagster_shared.check as check
import pytest
from dagster_shared.check import (
    CheckError,
    ElementCheckError,
    EvalContext,
    ImportFrom,
    NotImplementedCheckError,
    ParameterCheckError,
    build_check_call_str,
)
from dagster_shared.check.builder import INJECTED_CHECK_VAR
from dagster_shared.record import record

if TYPE_CHECKING:
    from dagster_shared.utils.test import TestType


@contextmanager
def raises_with_message(exc_type, message_text):
    with pytest.raises(exc_type) as exc_info:
        yield

    assert str(exc_info.value) == message_text


def is_python_three():
    return sys.version_info[0] >= 3


# ###################################################################################################
# ##### TYPE CHECKS
# ###################################################################################################

# ########################
# ##### BOOL
# ########################


def test_bool_param():
    assert check.bool_param(True, "b") is True
    assert check.bool_param(False, "b") is False

    with pytest.raises(ParameterCheckError):
        check.bool_param(None, "b")

    with pytest.raises(ParameterCheckError):
        check.bool_param(0, "b")

    with pytest.raises(ParameterCheckError):
        check.bool_param("val", "b")


def test_opt_bool_param():
    assert check.opt_bool_param(True, "b") is True
    assert check.opt_bool_param(False, "b") is False
    assert check.opt_bool_param(None, "b") is None
    assert check.opt_bool_param(None, "b", True) is True
    assert check.opt_bool_param(None, "b", False) is False

    with pytest.raises(ParameterCheckError):
        check.opt_bool_param(0, "b")

    with pytest.raises(ParameterCheckError):
        check.opt_bool_param("val", "b")


def test_bool_elem():
    ddict = {"a_true": True, "a_str": "a", "a_num": 1, "a_none": None}

    assert check.bool_elem(ddict, "a_true") is True

    with pytest.raises(ElementCheckError):
        check.bool_elem(ddict, "a_none")

    with pytest.raises(ElementCheckError):
        check.bool_elem(ddict, "a_num")

    with pytest.raises(ElementCheckError):
        check.bool_elem(ddict, "a_str")


def test_opt_bool_elem():
    ddict = {"a_bool": True, "a_float": 1.0, "a_int": 1, "a_none": None, "a_str": "a"}

    assert check.opt_bool_elem(ddict, "a_bool") is True
    assert check.opt_bool_elem(ddict, "a_none") is None
    assert check.opt_bool_elem(ddict, "nonexistentkey") is None

    for key in ["a_float", "a_int", "a_str"]:
        with pytest.raises(ElementCheckError):
            assert check.opt_bool_elem(ddict, key) is False


# ########################
# ##### CALLABLE
# ########################


def test_callable_param():
    lamb = lambda: 1
    assert check.callable_param(lamb, "lamb") == lamb

    with pytest.raises(ParameterCheckError):
        check.callable_param(None, "lamb")  # pyright: ignore[reportArgumentType]

    with pytest.raises(ParameterCheckError):
        check.callable_param(2, "lamb")  # pyright: ignore[reportArgumentType]


def test_opt_callable_param():
    lamb = lambda: 1
    assert check.opt_callable_param(lamb, "lamb") == lamb
    assert check.opt_callable_param(None, "lamb") is None
    assert check.opt_callable_param(None, "lamb") is None
    assert check.opt_callable_param(None, "lamb", default=lamb) == lamb

    with pytest.raises(ParameterCheckError):
        check.opt_callable_param(2, "lamb")  # pyright: ignore[reportCallIssue,reportArgumentType]


def test_is_callable():
    def fn():
        pass

    assert check.is_callable(fn) == fn
    assert check.is_callable(lambda: None)
    assert check.is_callable(lambda: None, "some desc")

    with pytest.raises(CheckError):
        check.is_callable(None)  # pyright: ignore[reportArgumentType]

    with pytest.raises(CheckError):
        check.is_callable(1)  # pyright: ignore[reportArgumentType]

    with pytest.raises(CheckError, match="some other desc"):
        check.is_callable(1, "some other desc")  # pyright: ignore[reportArgumentType]


# ########################
# ##### CLASS
# ########################


def test_opt_class_param():
    class Foo:
        pass

    assert check.opt_class_param(int, "foo")
    assert check.opt_class_param(Foo, "foo")

    assert check.opt_class_param(None, "foo") is None
    assert check.opt_class_param(None, "foo", Foo) is Foo

    with pytest.raises(CheckError):
        check.opt_class_param(check, "foo")

    with pytest.raises(CheckError):
        check.opt_class_param(234, "foo")

    with pytest.raises(CheckError):
        check.opt_class_param("bar", "foo")

    with pytest.raises(CheckError):
        check.opt_class_param(Foo(), "foo")


def test_class_param():
    class Bar:
        pass

    assert check.class_param(int, "foo")
    assert check.class_param(Bar, "foo")

    with pytest.raises(CheckError):
        check.class_param(None, "foo")  # pyright: ignore[reportArgumentType]

    with pytest.raises(CheckError):
        check.class_param(check, "foo")

    with pytest.raises(CheckError):
        check.class_param(234, "foo")  # pyright: ignore[reportArgumentType]

    with pytest.raises(CheckError):
        check.class_param("bar", "foo")  # pyright: ignore[reportArgumentType]

    with pytest.raises(CheckError):
        check.class_param(Bar(), "foo")  # pyright: ignore[reportArgumentType]

    class Super:
        pass

    class Sub(Super):
        pass

    class Alone:
        pass

    assert check.class_param(Sub, "foo", superclass=Super)

    with pytest.raises(CheckError):
        assert check.class_param(Alone, "foo", superclass=Super)

    with pytest.raises(CheckError):
        assert check.class_param("value", "foo", superclass=Super)  # pyright: ignore[reportArgumentType]

    assert check.opt_class_param(Sub, "foo", superclass=Super)
    assert check.opt_class_param(None, "foo", superclass=Super) is None

    with pytest.raises(CheckError):
        assert check.opt_class_param(Alone, "foo", superclass=Super)

    with pytest.raises(CheckError):
        assert check.opt_class_param("value", "foo", superclass=Super)


# ########################
# ##### DICT
# ########################


class Wrong:
    pass


class AlsoWrong:
    pass


DICT_TEST_CASES = [
    (dict(obj={}), True),
    (dict(obj={"a": 2}), True),
    (dict(obj=None), False),
    (dict(obj=0), False),
    (dict(obj=1), False),
    (dict(obj="foo"), False),
    (dict(obj=["foo"]), False),
    (dict(obj=[]), False),
    (dict(obj={"str": 1}, key_type=str, value_type=int), True),
    (dict(obj={"str": 1}, value_type=int), True),
    (dict(obj={"str": 1}, key_type=str), True),
    (dict(obj={"str": 1}), True),
    (dict(obj={}, key_type=str, value_type=int), True),
    (dict(obj={}, value_type=int), True),
    (dict(obj={}, key_type=str), True),
    (dict(obj={}), True),
    (
        dict(
            obj={"str": 1, "str2": "str", 1: "str", 2: "str"},
            key_type=(str, int),
            value_type=(str, int),
        ),
        True,
    ),
    (dict(obj={"str": 1}, key_type=Wrong, value_type=Wrong), False),
    (dict(obj={"str": 1}, key_type=Wrong, value_type=int), False),
    (dict(obj={"str": 1}, key_type=str, value_type=Wrong), False),
    (dict(obj={"str": 1}, key_type=Wrong), False),
    (dict(obj={"str": 1}, key_type=(Wrong, AlsoWrong)), False),
    (dict(obj={"str": 1}, value_type=(Wrong, AlsoWrong)), False),
]


@pytest.mark.parametrize("kwargs, should_succeed", DICT_TEST_CASES)
def test_dict_param(kwargs, should_succeed):
    if should_succeed:
        assert check.dict_param(**kwargs, param_name="name") == kwargs["obj"]
    else:
        with pytest.raises(CheckError):
            check.dict_param(**kwargs, param_name="name")


@pytest.mark.parametrize("kwargs, should_succeed", DICT_TEST_CASES)
def test_is_dict(kwargs, should_succeed):
    if should_succeed:
        assert check.is_dict(**kwargs) == kwargs["obj"]
    else:
        with pytest.raises(CheckError):
            check.is_dict(**kwargs)


def test_opt_dict_param_with_type():
    str_to_int = {"str": 1}
    assert check.opt_dict_param(str_to_int, "str_to_int", key_type=str, value_type=int)
    assert check.opt_dict_param(str_to_int, "str_to_int", value_type=int)
    assert check.opt_dict_param(str_to_int, "str_to_int", key_type=str)
    assert check.opt_dict_param(str_to_int, "str_to_int")

    assert check.opt_dict_param({}, "str_to_int", key_type=str, value_type=int) == {}
    assert check.opt_dict_param({}, "str_to_int", value_type=int) == {}
    assert check.opt_dict_param({}, "str_to_int", key_type=str) == {}
    assert check.opt_dict_param({}, "str_to_int") == {}

    assert check.opt_dict_param(None, "str_to_int", key_type=str, value_type=int) == {}
    assert check.opt_dict_param(None, "str_to_int", value_type=int) == {}
    assert check.opt_dict_param(None, "str_to_int", key_type=str) == {}
    assert check.opt_dict_param(None, "str_to_int") == {}

    assert check.opt_dict_param(
        {"str": 1, "str2": "str", 1: "str", 2: "str"},
        "multi_type_dict",
        key_type=(str, int),
        value_type=(str, int),
    )

    class Wrong:
        pass

    with pytest.raises(CheckError):
        assert check.opt_dict_param(str_to_int, "str_to_int", key_type=Wrong, value_type=Wrong)

    with pytest.raises(CheckError):
        assert check.opt_dict_param(str_to_int, "str_to_int", key_type=Wrong, value_type=int)

    with pytest.raises(CheckError):
        assert check.opt_dict_param(str_to_int, "str_to_int", key_type=str, value_type=Wrong)

    with pytest.raises(CheckError):
        assert check.opt_dict_param(str_to_int, "str_to_int", key_type=Wrong)

    with pytest.raises(CheckError):
        assert check.opt_dict_param(str_to_int, "str_to_int", value_type=Wrong)

    class AlsoWrong:
        pass

    with pytest.raises(CheckError):
        assert check.dict_param(str_to_int, "str_to_int", key_type=(Wrong, AlsoWrong))

    with pytest.raises(CheckError):
        assert check.dict_param(str_to_int, "str_to_int", value_type=(Wrong, AlsoWrong))


def test_opt_dict_param():
    assert check.opt_dict_param(None, "opt_dict_param") == {}
    assert check.opt_dict_param({}, "opt_dict_param") == {}
    ddict = {"a": 2}
    assert check.opt_dict_param(ddict, "opt_dict_param") == ddict

    with pytest.raises(ParameterCheckError):
        check.opt_dict_param(0, "opt_dict_param")  # pyright: ignore[reportArgumentType]

    with pytest.raises(ParameterCheckError):
        check.opt_dict_param(1, "opt_dict_param")  # pyright: ignore[reportArgumentType]

    with pytest.raises(ParameterCheckError):
        check.opt_dict_param("foo", "opt_dict_param")  # pyright: ignore[reportArgumentType]

    with pytest.raises(ParameterCheckError):
        check.opt_dict_param(["foo"], "opt_dict_param")  # pyright: ignore[reportArgumentType]

    with pytest.raises(ParameterCheckError):
        check.opt_dict_param([], "opt_dict_param")  # pyright: ignore[reportArgumentType]


def test_opt_nullable_dict_param():
    assert check.opt_nullable_dict_param(None, "opt_nullable_dict_param") is None
    assert check.opt_nullable_dict_param({}, "opt_nullable_dict_param") == {}
    ddict = {"a": 2}
    assert check.opt_nullable_dict_param(ddict, "opt_nullable_dict_param") == ddict

    with pytest.raises(ParameterCheckError):
        check.opt_nullable_dict_param(1, "opt_nullable_dict_param")

    with pytest.raises(ParameterCheckError):
        check.opt_nullable_dict_param("foo", "opt_nullable_dict_param")


def test_opt_two_dim_dict_param():
    assert check.opt_two_dim_dict_param({}, "foo") == {}
    assert check.opt_two_dim_dict_param({"key": {}}, "foo")
    assert check.opt_two_dim_dict_param({"key": {"key2": 2}}, "foo")
    assert check.opt_two_dim_dict_param(None, "foo") == {}

    with pytest.raises(CheckError):
        assert check.opt_two_dim_dict_param("str", "foo")


def test_two_dim_dict():
    assert check.two_dim_dict_param({}, "foo") == {}
    assert check.two_dim_dict_param({"key": {}}, "foo")
    assert check.two_dim_dict_param({"key": {"key2": 2}}, "foo")

    # make sure default dict passes
    default_dict = defaultdict(dict)
    default_dict["key"]["key2"] = 2
    assert check.two_dim_dict_param(default_dict, "foo")

    with raises_with_message(
        CheckError, """Param "foo" is not a dict. Got None which is type <class 'NoneType'>."""
    ):
        check.two_dim_dict_param(None, "foo")

    with raises_with_message(
        CheckError,
        "Value in dict mismatches expected type for key int_value. Expected value "
        "of type <class 'dict'>. Got value 2 of type <class 'int'>.",
    ):
        check.two_dim_dict_param({"int_value": 2}, "foo")

    with raises_with_message(
        CheckError,
        "Value in dict mismatches expected type for key level_two_value_mismatch. "
        "Expected value of type <class 'str'>. Got value 2 of type <class 'int'>.",
    ):
        check.two_dim_dict_param(
            {"level_one_key": {"level_two_value_mismatch": 2}}, "foo", value_type=str
        )

    with raises_with_message(
        CheckError,
        (
            "Key in dict mismatches type. Expected <class 'int'>. Got 'key'"
            if is_python_three()
            else "Key in dictionary mismatches type. Expected <type 'int'>. Got 'key'"
        ),
    ):
        assert check.two_dim_dict_param({"key": {}}, "foo", key_type=int)

    with raises_with_message(
        CheckError,
        (
            "Key in dict mismatches type. Expected <class 'int'>. Got 'level_two_key'"
            if is_python_three()
            else "Key in dictionary mismatches type. Expected <type 'int'>. Got 'level_two_key'"
        ),
    ):
        assert check.two_dim_dict_param({1: {"level_two_key": "something"}}, "foo", key_type=int)


def test_dict_elem():
    dict_value = {"blah": "blahblah"}
    ddict = {"dictkey": dict_value, "stringkey": "A", "nonekey": None}

    assert check.dict_elem(ddict, "dictkey") == dict_value

    with pytest.raises(CheckError):
        check.dict_elem(ddict, "stringkey")

    with pytest.raises(CheckError):
        check.dict_elem(ddict, "nonekey")

    with pytest.raises(CheckError):
        check.dict_elem(ddict, "nonexistantkey")

    assert check.dict_elem(
        {"foo": {"str": 1, "str2": "str", 1: "str", 2: "str"}},
        "foo",
        key_type=(str, int),
        value_type=(str, int),
    )


def test_opt_dict_elem():
    dict_value = {"blah": "blahblah"}
    ddict = {"dictkey": dict_value, "stringkey": "A", "nonekey": None}

    assert check.opt_dict_elem(ddict, "dictkey") == dict_value
    assert check.opt_dict_elem(ddict, "dictkey", key_type=str, value_type=str) == dict_value
    assert check.opt_dict_elem(ddict, "nonekey") == {}
    assert check.opt_dict_elem(ddict, "nonexistantkey") == {}

    with pytest.raises(CheckError):
        check.opt_dict_elem(ddict, "stringkey")

    with pytest.raises(CheckError):
        check.opt_dict_elem(ddict, "dictkey", key_type=str, value_type=int)


# ########################
# ##### FLOAT
# ########################


def test_float_param():
    assert check.float_param(-1.0, "param_name") == -1.0
    assert check.float_param(0.0, "param_name") == 0.0
    assert check.float_param(1.1, "param_name") == 1.1

    with pytest.raises(ParameterCheckError):
        check.float_param(None, "param_name")

    with pytest.raises(ParameterCheckError):
        check.float_param("s", "param_name")

    with pytest.raises(ParameterCheckError):
        check.float_param(1, "param_name")

    with pytest.raises(ParameterCheckError):
        check.float_param(0, "param_name")


def test_opt_float_param():
    assert check.opt_float_param(-1.0, "param_name") == -1.0
    assert check.opt_float_param(0.0, "param_name") == 0.0
    assert check.opt_float_param(1.1, "param_name") == 1.1
    assert check.opt_float_param(None, "param_name") is None

    with pytest.raises(ParameterCheckError):
        check.opt_float_param("s", "param_name")


def test_float_elem():
    ddict = {"a_bool": True, "a_float": 1.0, "a_int": 1, "a_none": None, "a_str": "a"}

    assert check.float_elem(ddict, "a_float") == 1.0

    with pytest.raises(ElementCheckError):
        check.float_elem(ddict, "a_bool")

    with pytest.raises(ElementCheckError):
        check.float_elem(ddict, "a_int")

    with pytest.raises(ElementCheckError):
        check.float_elem(ddict, "a_none")

    with pytest.raises(ElementCheckError):
        check.float_elem(ddict, "a_str")


def test_opt_float_elem():
    ddict = {"a_bool": True, "a_float": 1.0, "a_int": 1, "a_none": None, "a_str": "a"}

    assert check.opt_float_elem(ddict, "a_float") == 1.0

    assert check.opt_float_elem(ddict, "a_none") is None

    assert check.opt_float_elem(ddict, "nonexistentkey") is None

    with pytest.raises(ElementCheckError):
        check.opt_float_elem(ddict, "a_bool")

    with pytest.raises(ElementCheckError):
        check.opt_float_elem(ddict, "a_int")

    with pytest.raises(ElementCheckError):
        check.opt_float_elem(ddict, "a_str")


# ########################
# ##### GENERATOR
# ########################


def test_generator_param():
    def _test_gen():
        yield 1

    assert check.generator_param(_test_gen(), "gen")

    gen = _test_gen()
    assert check.generator(gen)
    assert list(gen) == [1]
    assert check.generator(gen)
    assert list(gen) == []

    with pytest.raises(ParameterCheckError):
        assert check.generator_param(list(gen), "gen")  # pyright: ignore[reportArgumentType]

    with pytest.raises(ParameterCheckError):
        assert check.generator_param(None, "gen")  # pyright: ignore[reportArgumentType]

    with pytest.raises(ParameterCheckError):
        assert check.generator_param(_test_gen, "gen")  # pyright: ignore[reportArgumentType]


def test_opt_generator_param():
    def _test_gen():
        yield 1

    assert check.opt_generator_param(_test_gen(), "gen")

    assert check.opt_generator_param(None, "gen") is None

    with pytest.raises(ParameterCheckError):
        assert check.opt_generator_param(_test_gen, "gen")


def test_generator():
    def _test_gen():
        yield 1

    assert check.generator(_test_gen())

    gen = _test_gen()
    assert check.generator(gen)

    with pytest.raises(ParameterCheckError):
        assert check.generator(list(gen))

    with pytest.raises(ParameterCheckError):
        assert check.generator(None)

    with pytest.raises(ParameterCheckError):
        assert check.generator(_test_gen)


def test_opt_generator():
    def _test_gen():
        yield 1

    assert check.opt_generator(_test_gen())

    gen = _test_gen()
    assert check.opt_generator(gen)
    assert check.opt_generator(None) is None

    with pytest.raises(ParameterCheckError):
        assert check.opt_generator(list(gen))

    with pytest.raises(ParameterCheckError):
        assert check.opt_generator(_test_gen)


# ########################
# ##### INT
# ########################


def test_int_param():
    assert check.int_param(-1, "param_name") == -1
    assert check.int_param(0, "param_name") == 0
    assert check.int_param(1, "param_name") == 1

    with pytest.raises(ParameterCheckError):
        check.int_param(None, "param_name")

    with pytest.raises(ParameterCheckError):
        check.int_param("s", "param_name")


def test_int_value_param():
    assert check.int_value_param(-1, -1, "param_name") == -1
    with pytest.raises(ParameterCheckError):
        check.int_value_param(None, -1, "param_name")

    with pytest.raises(ParameterCheckError):
        check.int_value_param(1, 0, "param_name")


def test_opt_int_param():
    assert check.opt_int_param(-1, "param_name") == -1
    assert check.opt_int_param(0, "param_name") == 0
    assert check.opt_int_param(1, "param_name") == 1
    assert check.opt_int_param(None, "param_name") is None

    with pytest.raises(ParameterCheckError):
        check.opt_int_param("s", "param_name")


def test_int_elem():
    ddict = {"a_bool": True, "a_float": 1.0, "a_int": 1, "a_none": None, "a_str": "a"}

    assert check.int_elem(ddict, "a_int") == 1

    assert check.int_elem(ddict, "a_bool") is True

    with pytest.raises(ElementCheckError):
        check.int_elem(ddict, "a_float")

    with pytest.raises(ElementCheckError):
        check.int_elem(ddict, "a_none")

    with pytest.raises(ElementCheckError):
        check.int_elem(ddict, "a_str")


def test_opt_int_elem():
    ddict = {"a_bool": True, "a_float": 1.0, "a_int": 1, "a_none": None, "a_str": "a"}

    assert check.opt_int_elem(ddict, "a_int") == 1

    assert check.opt_int_elem(ddict, "a_none") is None

    assert check.opt_int_elem(ddict, "nonexistentkey") is None

    assert check.opt_int_elem(ddict, "a_bool") is True

    with pytest.raises(ElementCheckError):
        check.opt_int_elem(ddict, "a_float")

    with pytest.raises(ElementCheckError):
        check.opt_int_elem(ddict, "a_str")


# ########################
# ##### INST
# ########################


def test_inst():
    class Foo:
        pass

    class Bar:
        pass

    obj = Foo()

    assert check.inst(obj, Foo) == obj

    with pytest.raises(CheckError, match="not a Bar"):
        check.inst(Foo(), Bar)

    with pytest.raises(CheckError, match="Expected only a Bar"):
        check.inst(Foo(), Bar, "Expected only a Bar")

    with pytest.raises(CheckError, match=re.escape("not one of ['Bar', 'Foo']")):
        check.inst(1, (Foo, Bar))

    with pytest.raises(CheckError, match=re.escape("not one of ['Bar', 'Foo']")):
        check.inst(1, (Foo, Bar), additional_message="a desc")

    with pytest.raises(CheckError, match=re.escape("a desc")):
        check.inst(1, (Foo, Bar), additional_message="a desc")


def test_inst_param():
    class Foo:
        pass

    class Bar:
        pass

    class Baaz:
        pass

    obj = Foo()

    assert check.inst_param(obj, "obj", Foo) == obj

    with pytest.raises(ParameterCheckError, match="not a Bar"):
        check.inst_param(None, "obj", Bar)

    with pytest.raises(ParameterCheckError, match="not a Bar"):
        check.inst_param(Bar, "obj", Bar)

    with pytest.raises(ParameterCheckError, match="not a Bar"):
        check.inst_param(Foo(), "obj", Bar)

    with pytest.raises(ParameterCheckError, match=r"not one of \['Bar', 'Foo'\]"):
        check.inst_param(None, "obj", (Foo, Bar))

    with pytest.raises(ParameterCheckError, match=r"not one of \['Bar', 'Foo'\]"):
        check.inst_param(Baaz(), "obj", (Foo, Bar))


def test_opt_inst_param():
    class Foo:
        pass

    class Bar:
        pass

    class Baaz:
        pass

    obj = Foo()

    assert check.opt_inst_param(obj, "obj", Foo) == obj
    assert check.opt_inst_param(None, "obj", Foo) is None
    assert check.opt_inst_param(None, "obj", Bar) is None

    with pytest.raises(ParameterCheckError, match="not a Bar"):
        check.opt_inst_param(Bar, "obj", Bar)

    with pytest.raises(ParameterCheckError, match="not a Bar"):
        check.opt_inst_param(Foo(), "obj", Bar)

    # check defaults

    default_obj = Foo()

    assert check.opt_inst_param(None, "obj", Foo, default_obj) is default_obj

    assert check.opt_inst_param(None, "obj", (Foo, Bar)) is None

    with pytest.raises(ParameterCheckError, match=r"not one of \['Bar', 'Foo'\]"):
        check.inst_param(Baaz(), "obj", (Foo, Bar))


# ########################
# ##### LIST
# ########################


def test_list_param():
    assert check.list_param([], "list_param") == []

    assert check.list_param(["foo"], "list_param", of_type=str) == ["foo"]

    with pytest.raises(ParameterCheckError):
        check.list_param(None, "list_param")

    with pytest.raises(ParameterCheckError):
        check.list_param("3u4", "list_param")

    with pytest.raises(CheckError):
        check.list_param(["foo"], "list_param", of_type=int)


def test_typed_list_param():
    class Foo:
        pass

    class Bar:
        pass

    assert check.list_param([], "list_param", Foo) == []
    foo_list = [Foo()]
    assert check.list_param(foo_list, "list_param", Foo) == foo_list

    with pytest.raises(CheckError):
        check.list_param([Bar()], "list_param", Foo)

    with pytest.raises(CheckError):
        check.list_param([None], "list_param", Foo)


def test_opt_list_param():
    assert check.opt_list_param(None, "list_param") == []
    assert check.opt_list_param(None, "list_param", of_type=str) == []
    assert check.opt_list_param([], "list_param") == []
    obj_list = [1]
    assert check.list_param(obj_list, "list_param") == obj_list
    assert check.opt_list_param(["foo"], "list_param", of_type=str) == ["foo"]

    with pytest.raises(ParameterCheckError):
        check.opt_list_param(0, "list_param")

    with pytest.raises(ParameterCheckError):
        check.opt_list_param("", "list_param")

    with pytest.raises(ParameterCheckError):
        check.opt_list_param("3u4", "list_param")

    with pytest.raises(CheckError):
        check.opt_list_param(["foo"], "list_param", of_type=int)


def test_opt_typed_list_param():
    class Foo:
        pass

    class Bar:
        pass

    assert check.opt_list_param(None, "list_param", Foo) == []
    assert check.opt_list_param([], "list_param", Foo) == []
    foo_list = [Foo()]
    assert check.opt_list_param(foo_list, "list_param", Foo) == foo_list

    with pytest.raises(CheckError):
        check.opt_list_param([Bar()], "list_param", Foo)

    with pytest.raises(CheckError):
        check.opt_list_param([None], "list_param", Foo)


def test_opt_nullable_list_param():
    assert check.opt_nullable_list_param(None, "list_param") is None
    assert check.opt_nullable_list_param([], "list_param") == []
    obj_list = [1]
    assert check.opt_nullable_list_param(obj_list, "list_param") == obj_list

    with pytest.raises(ParameterCheckError):
        check.opt_nullable_list_param(0, "list_param")  # pyright: ignore[reportCallIssue,reportArgumentType]

    with pytest.raises(ParameterCheckError):
        check.opt_nullable_list_param("", "list_param")  # pyright: ignore[reportCallIssue,reportArgumentType]

    with pytest.raises(ParameterCheckError):
        check.opt_nullable_list_param("3u4", "list_param")  # pyright: ignore[reportCallIssue,reportArgumentType]


def test_typed_is_list():
    class Foo:
        pass

    class Bar:
        pass

    assert check.is_list([], Foo) == []
    foo_list = [Foo()]
    assert check.is_list(foo_list, of_type=Foo) == foo_list

    assert check.is_list([Foo(), Bar()], of_type=(Foo, Bar))

    with pytest.raises(CheckError):
        check.is_list([Bar()], of_type=Foo)

    with pytest.raises(CheckError):
        check.is_list([None], of_type=Foo)

    with pytest.raises(CheckError):
        check.is_list([Foo(), Bar(), ""], of_type=(Foo, Bar))


def test_two_dim_list_param():
    assert check.two_dim_list_param([[1, 2], [2, 3]], "something")

    with pytest.raises(CheckError):
        assert check.two_dim_list_param(None, "something")

    with pytest.raises(CheckError):
        assert check.two_dim_list_param([1, 2, 4], "something")

    with pytest.raises(CheckError):
        assert check.two_dim_list_param([], "something")

    with pytest.raises(CheckError):
        assert check.two_dim_list_param([[1, 2], 3], "soemthing")

    with pytest.raises(CheckError):
        assert check.two_dim_list_param([[1, 2], [3.0, 4.1]], "something", of_type=int)

    with pytest.raises(CheckError):
        assert check.two_dim_list_param([[1, 2], [2, 3, 4]], "something")


def test_list_elem():
    list_value = ["blah", "blahblah"]
    ddict = {"listkey": list_value, "stringkey": "A", "nonekey": None}

    assert check.list_elem(ddict, "listkey") == list_value
    assert check.list_elem(ddict, "listkey", of_type=str) == list_value

    with pytest.raises(CheckError):
        assert check.list_elem(ddict, "nonekey") == []

    with pytest.raises(CheckError):
        assert check.list_elem(ddict, "nonexistantkey") == []

    with pytest.raises(CheckError):
        check.list_elem(ddict, "stringkey")

    with pytest.raises(CheckError):
        check.list_elem(ddict, "listkey", of_type=int)


def test_opt_list_elem():
    list_value = ["blah", "blahblah"]
    ddict = {"listkey": list_value, "stringkey": "A", "nonekey": None}

    assert check.opt_list_elem(ddict, "listkey") == list_value
    assert check.opt_list_elem(ddict, "listkey", of_type=str) == list_value
    assert check.opt_list_elem(ddict, "nonekey") == []
    assert check.opt_list_elem(ddict, "nonexistantkey") == []

    with pytest.raises(CheckError):
        check.opt_list_elem(ddict, "stringkey")

    with pytest.raises(CheckError):
        check.opt_list_elem(ddict, "listkey", of_type=int)


# ########################
# ##### MAPPING
# ########################


class SimpleMapping(collections.abc.Mapping):
    def __init__(self, **kwargs):
        self._dict = dict()
        for key, value in kwargs.items():
            self._dict[key] = value

    def __getitem__(self, key):
        return self._dict[key]

    def __iter__(self):
        return iter(self._dict)

    def __len__(self):
        return len(self._dict)


MAPPING_TEST_CASES = DICT_TEST_CASES + [
    (dict(obj=SimpleMapping(x=1), key_type=str, value_type=int), True),
]


@pytest.mark.parametrize("kwargs, should_succeed", MAPPING_TEST_CASES)
def test_mapping_param(kwargs, should_succeed):
    if should_succeed:
        assert check.mapping_param(**kwargs, param_name="name") == kwargs["obj"]
    else:
        with pytest.raises(CheckError):
            check.mapping_param(**kwargs, param_name="name")


def test_opt_mapping_param():
    mapping = SimpleMapping(x=1)
    assert check.opt_mapping_param(mapping, param_name="name") == mapping
    assert check.opt_mapping_param(mapping, param_name="name", key_type=str) == mapping
    assert check.opt_mapping_param(mapping, param_name="name", value_type=int) == mapping
    assert check.opt_mapping_param(None, param_name="name") == dict()

    with pytest.raises(CheckError):
        check.opt_mapping_param("foo", param_name="name")  # pyright: ignore[reportArgumentType]
    assert check.opt_nullable_mapping_param(None, "name") is None


# ########################
# ##### NOT NONE
# ########################


def test_not_none_param():
    assert check.not_none_param(1, "fine")
    check.not_none_param(0, "zero is fine")
    check.not_none_param("", "empty str is fine")

    with pytest.raises(CheckError):
        check.not_none_param(None, "none fails")


# ########################
# ##### PATH
# ########################


def test_path_param():
    from pathlib import Path

    assert check.path_param("/a/b.csv", "path_param") == "/a/b.csv"
    if sys.platform.startswith("win32"):
        assert check.opt_path_param(Path("c:\\a\\b.csv"), "path_param") == "c:\\a\\b.csv"
    else:
        assert check.opt_path_param(Path("/a/b.csv"), "path_param") == "/a/b.csv"

    with pytest.raises(ParameterCheckError):
        check.path_param(None, "path_param")  # pyright: ignore[reportArgumentType]

    with pytest.raises(ParameterCheckError):
        check.path_param(0, "path_param")  # pyright: ignore[reportArgumentType]


def test_opt_path_param():
    from pathlib import Path

    assert check.opt_path_param("/a/b.csv", "path_param") == "/a/b.csv"
    if sys.platform.startswith("win32"):
        assert check.opt_path_param(Path("c:\\a\\b.csv"), "path_param") == "c:\\a\\b.csv"
    else:
        assert check.opt_path_param(Path("/a/b.csv"), "path_param") == "/a/b.csv"
    assert check.opt_path_param(None, "path_param") is None

    with pytest.raises(ParameterCheckError):
        check.opt_path_param(0, "path_param")  # pyright: ignore[reportCallIssue,reportArgumentType]


# ########################
# ##### SET
# ########################


def test_set_param():
    assert check.set_param(set(), "set_param") == set()
    assert check.set_param(frozenset(), "set_param") == set()

    with pytest.raises(ParameterCheckError):
        check.set_param(None, "set_param")  # pyright: ignore[reportArgumentType]

    with pytest.raises(ParameterCheckError):
        check.set_param("3u4", "set_param")  # pyright: ignore[reportArgumentType]

    obj_set = {1}
    assert check.set_param(obj_set, "set_param") == obj_set

    obj_set_two = {1, 1, 2}
    obj_set_two_deduped = {1, 2}
    assert check.set_param(obj_set_two, "set_param") == obj_set_two_deduped
    assert check.set_param(obj_set_two, "set_param", of_type=int) == obj_set_two_deduped

    with pytest.raises(CheckError, match="Did you pass a class"):
        check.set_param({str}, "set_param", of_type=int)

    with pytest.raises(CheckError, match="Member of set mismatches type"):
        check.set_param({"foo"}, "set_param", of_type=int)


def test_opt_set_param():
    assert check.opt_set_param(None, "set_param") == set()
    assert check.opt_set_param(set(), "set_param") == set()
    assert check.opt_set_param(frozenset(), "set_param") == set()
    assert check.opt_set_param({3}, "set_param") == {3}

    with pytest.raises(ParameterCheckError):
        check.opt_set_param(0, "set_param")  # pyright: ignore[reportArgumentType]

    with pytest.raises(ParameterCheckError):
        check.opt_set_param("3u4", "set_param")  # pyright: ignore[reportArgumentType]


# ########################
# ##### SEQUENCE
# ########################


@record
class SomeRecord: ...


def test_sequence_param():
    assert check.sequence_param([], "sequence_param") == []
    assert check.sequence_param(tuple(), "sequence_param") == tuple()

    assert check.sequence_param(["foo"], "sequence_param", of_type=str) == ["foo"]

    with pytest.raises(ParameterCheckError):
        check.sequence_param(None, "sequence_param")  # pyright: ignore[reportArgumentType]

    with pytest.raises(CheckError):
        check.sequence_param(1, "sequence_param", of_type=int)  # pyright: ignore[reportArgumentType]

    with pytest.raises(CheckError):
        check.sequence_param(["foo"], "sequence_param", of_type=int)

    with pytest.raises(CheckError):
        check.sequence_param("foo", "sequence_param")

    with pytest.raises(CheckError, match="str is a disallowed Sequence type"):
        check.sequence_param("foo", "sequence_param", of_type=str)

    with pytest.raises(CheckError):
        check.sequence_param(SomeRecord(), "sequence_param")  # pyright: ignore[reportArgumentType]


def test_opt_sequence_param():
    assert check.opt_sequence_param([], "sequence_param") == []
    assert check.opt_sequence_param(tuple(), "sequence_param") == tuple()

    assert check.opt_sequence_param(["foo"], "sequence_param", of_type=str) == ["foo"]

    assert check.opt_sequence_param(None, "sequence_param") == []

    with pytest.raises(CheckError):
        check.opt_sequence_param(1, "sequence_param", of_type=int)  # pyright: ignore[reportArgumentType]

    with pytest.raises(CheckError):
        check.opt_sequence_param(["foo"], "sequence_param", of_type=int)

    with pytest.raises(CheckError):
        check.opt_sequence_param("foo", "sequence_param")

    with pytest.raises(CheckError, match="str is a disallowed Sequence type"):
        check.opt_sequence_param("foo", "sequence_param", of_type=str)

    with pytest.raises(CheckError):
        check.opt_sequence_param(SomeRecord(), "sequence_param")  # pyright: ignore[reportArgumentType]


def test_opt_nullable_sequence_param():
    assert check.opt_nullable_sequence_param([], "sequence_param") == []
    assert check.opt_nullable_sequence_param(tuple(), "sequence_param") == tuple()

    assert check.opt_nullable_sequence_param(["foo"], "sequence_param", of_type=str) == ["foo"]

    assert check.opt_nullable_sequence_param(None, "sequence_param") is None

    with pytest.raises(CheckError):
        check.opt_nullable_sequence_param(1, "sequence_param", of_type=int)  # pyright: ignore[reportCallIssue,reportArgumentType]

    with pytest.raises(CheckError):
        check.opt_nullable_sequence_param(["foo"], "sequence_param", of_type=int)

    with pytest.raises(CheckError, match="str is a disallowed Sequence type"):
        assert check.opt_nullable_sequence_param("foo", "sequence_param", of_type=str)

    with pytest.raises(CheckError):
        check.opt_nullable_sequence_param(SomeRecord(), "sequence_param")  # pyright: ignore[reportCallIssue,reportArgumentType]


# ########################
# ##### STR
# ########################


def test_str_param():
    assert check.str_param("a", "str_param") == "a"
    assert check.str_param("", "str_param") == ""
    assert check.str_param("a", "unicode_param") == "a"

    with pytest.raises(ParameterCheckError):
        check.str_param(None, "str_param")

    with pytest.raises(ParameterCheckError):
        check.str_param(0, "str_param")

    with pytest.raises(ParameterCheckError):
        check.str_param(1, "str_param")


def test_opt_str_param():
    assert check.opt_str_param("a", "str_param") == "a"
    assert check.opt_str_param("", "str_param") == ""
    assert check.opt_str_param("a", "unicode_param") == "a"
    assert check.opt_str_param(None, "str_param") is None
    assert check.opt_str_param(None, "str_param", "foo") == "foo"

    with pytest.raises(ParameterCheckError):
        check.opt_str_param(0, "str_param")

    with pytest.raises(ParameterCheckError):
        check.opt_str_param(1, "str_param")


def test_opt_nonempty_str_param():
    assert check.opt_nonempty_str_param("a", "str_param") == "a"
    assert check.opt_nonempty_str_param("", "str_param") is None
    assert check.opt_nonempty_str_param("", "str_param", "foo") == "foo"
    assert check.opt_nonempty_str_param("a", "unicode_param") == "a"
    assert check.opt_nonempty_str_param(None, "str_param") is None
    assert check.opt_nonempty_str_param(None, "str_param", "foo") == "foo"

    with pytest.raises(ParameterCheckError):
        check.opt_nonempty_str_param(0, "str_param")

    with pytest.raises(ParameterCheckError):
        check.opt_nonempty_str_param(1, "str_param")


def test_string_elem():
    ddict = {"a_str": "a", "a_num": 1, "a_none": None}

    assert check.str_elem(ddict, "a_str") == "a"

    with pytest.raises(ElementCheckError):
        assert check.str_elem(ddict, "a_none")

    with pytest.raises(ElementCheckError):
        check.str_elem(ddict, "a_num")


def test_opt_string_elem():
    ddict = {"a_str": "a", "a_num": 1, "a_none": None}

    assert check.opt_str_elem(ddict, "a_str") == "a"

    assert check.opt_str_elem(ddict, "a_none") is None

    assert check.opt_str_elem(ddict, "nonexistentkey") is None

    with pytest.raises(ElementCheckError):
        check.opt_str_elem(ddict, "a_num")


# ########################
# ##### TUPLE
# ########################


def test_tuple_param():
    assert check.tuple_param((1, 2), "something")

    with pytest.raises(CheckError):
        assert check.tuple_param(None, "something")  # pyright: ignore[reportArgumentType]

    with pytest.raises(CheckError):
        assert check.tuple_param(1, "something")  # pyright: ignore[reportArgumentType]

    with pytest.raises(CheckError):
        assert check.tuple_param([1], "something")  # pyright: ignore[reportArgumentType]

    with pytest.raises(CheckError):
        assert check.tuple_param({1: 2}, "something")  # pyright: ignore[reportArgumentType]

    with pytest.raises(CheckError):
        assert check.tuple_param("kdjfkd", "something")  # pyright: ignore[reportArgumentType]

    assert check.tuple_param((3, 4), "something", of_type=int)
    assert check.tuple_param(("foo", "bar"), "something", of_type=str)

    assert check.tuple_param((3, 4), "something", of_type=(int,))
    assert check.tuple_param((3, 4), "something", of_type=(int, str))
    assert check.tuple_param((3, "bar"), "something", of_type=(int, str))

    with pytest.raises(CheckError):
        check.tuple_param((3, 4, 5), "something", of_type=str)

    with pytest.raises(CheckError):
        check.tuple_param((3, 4), "something", of_type=(str,))

    assert check.tuple_param((3, "a"), "something", of_shape=(int, str))

    with pytest.raises(CheckError):
        check.tuple_param((3, "a"), "something", of_shape=(int, str, int))

    with pytest.raises(CheckError):
        check.tuple_param((3, "a"), "something", of_shape=(str, int))

    with pytest.raises(CheckError):
        check.is_tuple((3, 4), of_shape=(int, int), of_type=int)


def test_opt_tuple_param():
    assert check.opt_tuple_param((1, 2), "something")
    assert check.opt_tuple_param(None, "something") == tuple()

    with pytest.raises(CheckError):
        check.opt_tuple_param(1, "something")

    with pytest.raises(CheckError):
        check.opt_tuple_param([1], "something")

    with pytest.raises(CheckError):
        check.opt_tuple_param({1: 2}, "something")

    with pytest.raises(CheckError):
        check.opt_tuple_param("kdjfkd", "something")

    assert check.opt_tuple_param((3, 4), "something", of_type=int)
    assert check.opt_tuple_param(("foo", "bar"), "something", of_type=str)

    assert check.opt_tuple_param((3, 4), "something", of_type=(int,))
    assert check.opt_tuple_param((3, 4), "something", of_type=(int, str))
    assert check.opt_tuple_param((3, "bar"), "something", of_type=(int, str))

    with pytest.raises(CheckError):
        check.opt_tuple_param((3, 4, 5), "something", of_type=str)

    with pytest.raises(CheckError):
        check.opt_tuple_param((3, 4), "something", of_type=(str,))

    assert check.opt_tuple_param((3, "a"), "something", of_shape=(int, str))

    with pytest.raises(CheckError):
        check.opt_tuple_param((3, "a"), "something", of_shape=(int, str, int))

    with pytest.raises(CheckError):
        check.opt_tuple_param((3, "a"), "something", of_shape=(str, int))

    with pytest.raises(CheckError):
        check.opt_tuple_param((3, 4), "something", of_shape=(int, int), of_type=int)


def test_opt_nullable_tuple_param():
    assert check.opt_nullable_tuple_param((1, 2), "something")
    assert check.opt_nullable_tuple_param(None, "something") is None

    with pytest.raises(CheckError):
        check.opt_nullable_tuple_param([3, 4], "something", of_shape=(int, int), of_type=int)  # pyright: ignore[reportCallIssue,reportArgumentType]


def test_is_tuple():
    assert check.is_tuple(()) == ()

    with pytest.raises(CheckError):
        check.is_tuple(None)

    with pytest.raises(CheckError):
        check.is_tuple("3u4")

    with pytest.raises(CheckError, match="Did you pass a class"):
        check.is_tuple((str,), of_type=int)

    with pytest.raises(CheckError):
        check.is_tuple(SomeRecord())


def test_tuple_elem():
    tuple_value = ("blah", "blahblah")
    ddict = {"tuplekey": tuple_value, "stringkey": "A", "nonekey": None, "reckey": SomeRecord()}

    assert check.tuple_elem(ddict, "tuplekey") == tuple_value
    assert check.tuple_elem(ddict, "tuplekey", of_type=str) == tuple_value

    with pytest.raises(CheckError):
        check.tuple_elem(ddict, "nonekey")

    with pytest.raises(CheckError):
        check.tuple_elem(ddict, "nonexistantkey")

    with pytest.raises(CheckError):
        check.tuple_elem(ddict, "stringkey")

    with pytest.raises(CheckError):
        check.tuple_elem(ddict, "tuplekey", of_type=int)

    with pytest.raises(CheckError):
        check.tuple_elem(ddict, "reckey")


def test_opt_tuple_elem():
    tuple_value = ("blah", "blahblah")
    ddict = {"tuplekey": tuple_value, "stringkey": "A", "nonekey": None, "reckey": SomeRecord()}

    assert check.opt_tuple_elem(ddict, "tuplekey") == tuple_value
    assert check.opt_tuple_elem(ddict, "tuplekey", of_type=str) == tuple_value
    assert check.opt_tuple_elem(ddict, "nonekey") == tuple()
    assert check.opt_tuple_elem(ddict, "nonexistantkey") == tuple()

    with pytest.raises(CheckError):
        check.opt_tuple_elem(ddict, "stringkey")

    with pytest.raises(CheckError):
        check.opt_tuple_elem(ddict, "tuplekey", of_type=int)

    with pytest.raises(CheckError):
        check.opt_tuple_elem(ddict, "reckey")


def test_typed_is_tuple():
    class Foo:
        pass

    class Bar:
        pass

    assert check.is_tuple((), Foo) == ()
    foo_tuple = (Foo(),)
    assert check.is_tuple(foo_tuple, Foo) == foo_tuple
    assert check.is_tuple(foo_tuple, (Foo, Bar))

    with pytest.raises(CheckError):
        check.is_tuple((Bar(),), Foo)

    with pytest.raises(CheckError):
        check.is_tuple((None,), Foo)

    assert check.is_tuple((Foo(), Bar()), of_shape=(Foo, Bar))

    with pytest.raises(CheckError):
        check.is_tuple((Foo(),), of_shape=(Foo, Bar))

    with pytest.raises(CheckError):
        check.is_tuple((Foo(), Foo()), of_shape=(Foo, Bar))

    with pytest.raises(CheckError):
        check.is_tuple((Foo(), Foo()), of_shape=(Foo, Foo), of_type=Foo)


# ###################################################################################################
# ##### OTHER CHECKS
# ###################################################################################################


def test_param_invariant():
    check.param_invariant(True, "some_param")
    num_to_check = 1
    check.param_invariant(num_to_check == 1, "some_param")

    with pytest.raises(ParameterCheckError):
        check.param_invariant(num_to_check == 2, "some_param")

    with pytest.raises(ParameterCheckError):
        check.param_invariant(False, "some_param")

    with pytest.raises(ParameterCheckError):
        check.param_invariant(0, "some_param")

    check.param_invariant(1, "some_param")

    with pytest.raises(ParameterCheckError):
        check.param_invariant("", "some_param")

    check.param_invariant("1kjkjsf", "some_param")

    with pytest.raises(ParameterCheckError):
        check.param_invariant({}, "some_param")

    check.param_invariant({234: "1kjkjsf"}, "some_param")

    with pytest.raises(ParameterCheckError):
        check.param_invariant([], "some_param")

    check.param_invariant([234], "some_param")


def test_invariant():
    assert check.invariant(True)

    with pytest.raises(CheckError):
        check.invariant(False)

    with pytest.raises(CheckError, match="Some Unique String"):
        check.invariant(False, "Some Unique String")

    empty_list = []

    with pytest.raises(CheckError, match="Invariant failed"):
        check.invariant(empty_list)


def test_failed():
    with pytest.raises(CheckError, match="some desc"):
        check.failed("some desc")

    with pytest.raises(CheckError, match="must be a string"):
        check.failed(0)  # pyright: ignore[reportArgumentType]


def test_not_implemented():
    with pytest.raises(NotImplementedCheckError, match="some string"):
        check.not_implemented("some string")

    with pytest.raises(CheckError, match="desc argument must be a string"):
        check.not_implemented(None)  # pyright: ignore[reportArgumentType]


def test_iterable():
    assert check.iterable_param([], "thisisfine") == []
    assert check.iterable_param([1], "thisisfine") == [1]
    assert check.iterable_param([1], "thisisfine", of_type=int) == [1]
    assert check.iterable_param((i for i in [1, 2]), "thisisfine")
    # assert that it does not coerce generator to list
    assert check.iterable_param((i for i in [1, 2]), "thisisfine") != [1, 2]
    assert list(check.iterable_param((i for i in [1, 2]), "thisisfine")) == [1, 2]

    with pytest.raises(CheckError, match=r"Iterable.*str"):
        check.iterable_param("lkjsdkf", "stringisiterable")

    with pytest.raises(CheckError, match=r"Iterable.*None"):
        check.iterable_param(None, "nonenotallowed")  # pyright: ignore[reportArgumentType]

    with pytest.raises(CheckError, match=r"Iterable.*int"):
        check.iterable_param(1, "intnotallowed")  # pyright: ignore[reportArgumentType]

    with pytest.raises(CheckError, match="Member of iterable mismatches type"):
        check.iterable_param([1], "typemismatch", of_type=str)

    with pytest.raises(CheckError, match="Member of iterable mismatches type"):
        check.iterable_param(["atr", 2], "typemismatchmixed", of_type=str)

    with pytest.raises(CheckError, match="Member of iterable mismatches type"):
        check.iterable_param(["atr", None], "nonedoesntcount", of_type=str)

    with pytest.raises(CheckError):
        check.iterable_param(SomeRecord(), "nonenotallowed")  # pyright: ignore[reportArgumentType]


def test_opt_iterable():
    assert check.opt_iterable_param(None, "thisisfine") == []
    assert check.opt_iterable_param([], "thisisfine") == []
    assert check.opt_iterable_param([1], "thisisfine") == [1]
    assert check.opt_iterable_param((i for i in [1, 2]), "thisisfine")
    # assert that it does not coerce generator to list
    assert check.opt_iterable_param((i for i in [1, 2]), "thisisfine") != [1, 2]
    # not_none coerces to Iterable[T] so
    assert list(check.not_none(check.opt_iterable_param((i for i in [1, 2]), "thisisfine"))) == [
        1,
        2,
    ]

    check.opt_iterable_param(None, "noneisallowed")

    with pytest.raises(CheckError, match=r"Iterable.*str"):
        check.opt_iterable_param("lkjsdkf", "stringisiterable")

    with pytest.raises(CheckError, match=r"Iterable.*int"):
        check.opt_iterable_param(1, "intnotallowed")  # pyright: ignore[reportArgumentType]

    with pytest.raises(CheckError, match="Member of iterable mismatches type"):
        check.opt_iterable_param([1], "typemismatch", of_type=str)

    with pytest.raises(CheckError, match="Member of iterable mismatches type"):
        check.opt_iterable_param(["atr", 2], "typemismatchmixed", of_type=str)

    with pytest.raises(CheckError, match="Member of iterable mismatches type"):
        check.opt_iterable_param(["atr", None], "nonedoesntcount", of_type=str)

    with pytest.raises(CheckError):
        check.opt_iterable_param(SomeRecord(), "nonenotallowed")  # pyright: ignore[reportArgumentType]


def test_is_iterable() -> None:
    assert check.is_iterable([]) == []
    assert check.is_iterable((1, 2)) == tuple([1, 2])
    assert check.is_iterable("foo") == "foo"  # str is iterable
    assert check.is_iterable({"a": 1}) == {"a": 1}  # dict is iterable

    assert check.is_iterable([1, "str"]) == [1, "str"]

    with pytest.raises(CheckError):
        check.is_iterable([1, "str"], of_type=int)

    with pytest.raises(CheckError):
        check.is_iterable([1, "str"], of_type=str)

    with pytest.raises(CheckError):
        check.is_iterable(None)

    with pytest.raises(CheckError):
        check.is_iterable(1)


def test_is_iterable_typing() -> None:
    def returns_iterable_of_int_but_typed_any() -> Any:
        return [1, 2]

    def returns_iterable_of_t() -> Iterable[int]:
        any_typed = returns_iterable_of_int_but_typed_any()
        retval = check.is_iterable(any_typed, of_type=str)
        # That the type: ignore is necessary is proof that
        # is_iterable flows type information correctly
        return retval  # type: ignore

    # meaningless assert. The test is show the typechecker working
    assert returns_iterable_of_t


# ###################################################################################################
# ##### CHECK BUILDER
# ###################################################################################################


def build_check_call(ttype, name, eval_ctx: EvalContext):
    body = build_check_call_str(ttype, name, eval_ctx)
    lazy_import_str = "\n    ".join(
        f"from {module} import {t}" for t, module in eval_ctx.lazy_imports.items()
    )
    eval_ctx.local_ns[INJECTED_CHECK_VAR] = check
    fn = f"""
def _check({name}):
    {lazy_import_str}
    return {body}
"""
    return eval_ctx.compile_fn(fn, "_check")


class Foo: ...


class SubFoo(Foo): ...


class Bar: ...


T = TypeVar("T")


class Gen(Generic[T]): ...


class SubGen(Gen[str]): ...


class MyTypedDict(TypedDict):
    foo: str
    bar: str


BUILD_CASES = [
    (int, [4], ["4"]),
    (float, [4.2], ["4.1"]),
    (str, ["hi"], [Foo()]),
    (Bar, [Bar()], [Foo()]),
    (Bar | None, [Bar()], [Foo()]),
    (List[str], [["a", "b"]], [[1, 2]]),
    (Sequence[str], [["a", "b"]], [[1, 2], "just_a_string"]),
    (Iterable[str], [["a", "b"]], [[1, 2]]),
    (Set[str], [{"a", "b"}], [{1, 2}]),
    (AbstractSet[str], [{"a", "b"}], [{1, 2}]),
    (AbstractSet[str] | None, [{"a", "b"}, None], [{1, 2}]),
    (
        Mapping[str, AbstractSet[str]],
        [
            {"letters": {"a", "b"}},
            # should fail, but we do not yet handle inner collection types,
            # check.mapping_param(..., key_type=str, value_type=AbstractSet)
            {"numbers": {1, 2}},
        ],
        [
            {"letters": ["a", "b"]},
        ],
    ),
    (Dict[str, int], [{"a": 1}], [{1: "a"}]),
    (Mapping[str, int], [{"a": 1}], [{1: "a"}]),
    (int | None, [None], ["4"]),
    (Bar | None, [None], [Foo()]),
    (list[str] | None, [["a", "b"]], [[1, 2]]),
    (Sequence[str] | None, [["a", "b"]], [[1, 2]]),
    (Iterable[str] | None, [["a", "b"]], [[1, 2]]),
    (set[str] | None, [{"a", "b"}], [{1, 2}]),
    (dict[str, int] | None, [{"a": 1}], [{1: "a"}]),
    (Mapping[str, int] | None, [{"a": 1}], [{1: "a"}]),
    (Annotated[Bar, None], [Bar()], [Foo()]),
    (Annotated["Bar", None], [Bar()], [Foo()]),
    (List[Annotated[Bar, None]], [[Bar()], []], [[Foo()]]),
    (
        List[Annotated["TestType", ImportFrom("dagster_shared.utils.test")]],
        [[]],  # avoid importing TestType
        [[Foo()]],
    ),
    (Union[bool, Foo], [True], [None]),  # noqa: UP007
    (bool | Foo, [True], [None]),
    (Union[Foo, "Bar"], [Bar()], [None]),
    (TypeVar("T", bound=Foo), [Foo(), SubFoo()], [Bar()]),
    (TypeVar("T", bound=Foo | None), [None], [Bar()]),
    (TypeVar("T"), [Foo(), None], []),
    (Literal["apple"], ["apple"], ["banana"]),
    (Literal["apple", "manzana"], ["apple", "manzana"], ["banana"]),
    (Callable, [lambda x: x, int], [4]),
    (Callable[[], int], [lambda x: x, int], [4]),
    (
        Callable[[], int] | None,
        [
            None,
            lambda x: x,
        ],
        [4],
    ),
    # fwd refs
    ("Foo", [Foo()], [Bar()]),
    (Optional["Foo"], [Foo()], [Bar()]),
    (Mapping[str, Optional["Foo"]], [{"foo": Foo()}], [{"bar": Bar()}]),
    (Mapping[str, Optional["Foo"]], [{"foo": Foo()}], [{"bar": Bar()}]),
    (Gen, [Gen()], [Bar()]),
    (Gen[str], [Gen()], [Bar()]),
    (SubGen, [SubGen()], [Bar()]),
    (Sequence[SubGen], [[SubGen()]], [[Bar()]]),
    (Sequence[Gen[str]], [[Gen()]], [[Bar()]]),
    (
        MyTypedDict,
        [
            {"foo": "f", "bar": "b"},
            {},  # cant actually validate TypeDict structure, just that its a dict
        ],
        [None, Foo()],
    ),
    (MyTypedDict | None, [{"foo": "f", "bar": "b"}, None], [Foo()]),
]


@pytest.mark.parametrize("ttype, should_succeed, should_fail", BUILD_CASES)
def test_build_check_call(
    ttype: type, should_succeed: Sequence[object], should_fail: Sequence[object]
) -> None:
    eval_ctx = EvalContext(globals(), locals(), {})
    check_call = build_check_call(ttype, "test_param", eval_ctx)

    for obj in should_succeed:
        check_call(obj)

    for obj in should_fail:
        with pytest.raises(CheckError):
            check_call(obj)


def test_build_check_errors() -> None:
    with pytest.raises(CheckError, match=r"Unable to resolve ForwardRef\('NoExist'\)"):
        build_check_call(
            List["NoExist"],  # type: ignore # noqa
            "bad",
            EvalContext(globals(), locals(), {}),
        )


def test_forward_ref_flow() -> None:
    # original context captured at decl
    eval_ctx = EvalContext(globals(), locals(), {})
    ttype = List["Late"]  # class not yet defined

    class Late: ...

    with pytest.raises(CheckError):
        # can not build call since ctx was captured before definition
        build_check_call(ttype, "ok", eval_ctx)

    eval_ctx.update_from_frame(0)  # update from callsite frame
    # now it works
    call = build_check_call(ttype, "ok", eval_ctx)
    call([Late()])


def test_never():
    def _val() -> int | str:
        return 4

    v = _val()
    if isinstance(v, str):
        ...
    elif isinstance(v, int):
        ...
    else:
        check.assert_never(v)

    with pytest.raises(check.CheckError, match=r"Unhandled value: 4 \(int\)"):
        if isinstance(v, str):
            ...
        else:
            check.assert_never(v)  # type: ignore # good job type checker



import warnings
from collections.abc import Callable, Iterator
from contextlib import contextmanager
from contextvars import ContextVar
from typing import TypeVar

from dagster_shared import check

T = TypeVar("T")

_warnings_on = ContextVar("_warnings_on", default=True)

# ########################
# ##### PREVIEW
# ########################


class PreviewWarning(Warning):
    pass


def preview_warning(
    subject: str,
    additional_warn_text: str | None = None,
    stacklevel: int = 3,
):
    if not _warnings_on.get():
        return

    warnings.warn(
        f"{subject} is currently in preview, and may have breaking changes in patch version releases. "
        f"This feature is not considered ready for production use."
        + ((" " + additional_warn_text) if additional_warn_text else ""),
        category=PreviewWarning,
        stacklevel=stacklevel,
    )


# ########################
# ##### BETA
# ########################


class BetaWarning(Warning):
    pass


def beta_warning(
    subject: str,
    additional_warn_text: str | None = None,
    stacklevel: int = 3,
):
    if not _warnings_on.get():
        return

    warnings.warn(
        f"{subject} is currently in beta, and may have breaking changes in minor version releases, "
        f"with behavior changes in patch releases."
        + ((" " + additional_warn_text) if additional_warn_text else ""),
        category=BetaWarning,
        stacklevel=stacklevel,
    )


# ########################
# ##### SUPERSEDED
# ########################


class SupersessionWarning(FutureWarning):
    pass


def supersession_warning(
    subject: str,
    additional_warn_text: str | None = None,
    stacklevel: int = 3,
):
    if not _warnings_on.get():
        return

    warnings.warn(
        f"{subject} is superseded and its usage is discouraged."
        + ((" " + additional_warn_text) if additional_warn_text else ""),
        category=SupersessionWarning,
        stacklevel=stacklevel,
    )


# ########################
# ##### DEPRECATED
# ########################


def normalize_renamed_param(
    new_val: T,
    new_arg: str,
    old_val: T,
    old_arg: str,
    coerce_old_to_new: Callable[[T], T] | None = None,
) -> T:
    """Utility for managing backwards compatibility of a renamed parameter.

    .. code-block::

       # The name of param `old_flag` is being updated to `new_flag`, but we are temporarily
       # accepting either param.
       def is_new(old_flag=None, new_flag=None):
           return canonicalize_backcompat_args(
               new_val=new_flag,
               new_arg='new_flag',
               old_val=old_flag,
               old_arg='old_flag',
               breaking_version='0.9.0',
               coerce_old_to_new=lambda val: not val,
           )

    In the above example, if the caller sets both new_flag and old_flag, it will fail by throwing
    a CheckError. If the caller sets the new_flag, it's returned unaltered. If the caller sets
    old_flag, it will return the old_flag run through the coercion function.
    """
    check.str_param(new_arg, "new_arg")
    check.str_param(old_arg, "old_arg")
    check.opt_callable_param(coerce_old_to_new, "coerce_old_to_new")
    if new_val is not None and old_val is not None:
        check.failed(f'Do not use deprecated "{old_arg}" now that you are using "{new_arg}".')
    elif old_val is not None:
        return coerce_old_to_new(old_val) if coerce_old_to_new else old_val
    else:
        return new_val


def deprecation_warning(
    subject: str,
    breaking_version: str,
    additional_warn_text: str | None = None,
    stacklevel: int = 3,
):
    if not _warnings_on.get():
        return

    warnings.warn(
        f"{subject} is deprecated and will be removed in {breaking_version}."
        + ((" " + additional_warn_text) if additional_warn_text else ""),
        category=DeprecationWarning,
        stacklevel=stacklevel,
    )


# ########################
# ##### Config arg warning
# ########################

CONFIG_WARNING_HELP = (
    "To mute this warning, invoke"
    ' warnings.filterwarnings("ignore", category=dagster.ConfigArgumentWarning) or use'
    " one of the other methods described at"
    " https://docs.python.org/3/library/warnings.html#describing-warning-filters."
)


class ConfigArgumentWarning(SyntaxWarning):
    pass


def config_argument_warning(param_name: str, function_name: str) -> None:
    warnings.warn(
        f"Parameter '{param_name}' on op/asset function '{function_name}' was annotated as"
        " a dagster.Config type. Did you mean to name this parameter 'config'"
        " instead?\n\n"
        f"{CONFIG_WARNING_HELP}",
        ConfigArgumentWarning,
    )


# ########################
# ##### DISABLE DAGSTER WARNINGS
# ########################


@contextmanager
def disable_dagster_warnings() -> Iterator[None]:
    # If warnings are already disabled, do nothing. Nested resets of the token in finally blocks
    # have occasionally had inconsistent ordering in the past, which can lead to `_warnings_on`
    # being permanently turned off. By instantly returning, we prevent this nesting.
    if not _warnings_on.get():
        yield
    else:
        token = None
        try:
            token = _warnings_on.set(False)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=DeprecationWarning)
                warnings.simplefilter("ignore", category=SupersessionWarning)
                warnings.simplefilter("ignore", category=PreviewWarning)
                warnings.simplefilter("ignore", category=BetaWarning)
                yield
        finally:
            if token is not None:
                _warnings_on.reset(token)



import time
from collections.abc import Mapping
from enum import Enum
from typing import Any, Optional

from dagster_shared.dagster_model import DagsterModel
from dagster_shared.serdes import whitelist_for_serdes

LOCAL_STATE_VERSION = "__local__"
CODE_SERVER_STATE_VERSION = "__code_server__"


class DefsStateManagementType(Enum):
    VERSIONED_STATE_STORAGE = "VERSIONED_STATE_STORAGE"
    LOCAL_FILESYSTEM = "LOCAL_FILESYSTEM"
    LEGACY_CODE_SERVER_SNAPSHOTS = "LEGACY_CODE_SERVER_SNAPSHOTS"


@whitelist_for_serdes
class DefsKeyStateInfo(DagsterModel):
    """Records information about the version of the state for a given defs key."""

    version: str
    create_timestamp: float

    @classmethod
    def from_graphql(cls, data: dict[str, Any]) -> "DefsKeyStateInfo":
        return cls(version=data["version"], create_timestamp=data["createTimestamp"])

    @property
    def management_type(self) -> DefsStateManagementType:
        if self.version == LOCAL_STATE_VERSION:
            return DefsStateManagementType.LOCAL_FILESYSTEM
        elif self.version == CODE_SERVER_STATE_VERSION:
            return DefsStateManagementType.LEGACY_CODE_SERVER_SNAPSHOTS
        else:
            return DefsStateManagementType.VERSIONED_STATE_STORAGE


@whitelist_for_serdes
class DefsStateInfo(DagsterModel):
    """All of the information about the state version that will be used to load a given code location."""

    info_mapping: Mapping[str, DefsKeyStateInfo | None]

    @staticmethod
    def empty() -> "DefsStateInfo":
        return DefsStateInfo(info_mapping={})

    @staticmethod
    def add_version(
        current_info: Optional["DefsStateInfo"],
        key: str,
        version: str | None,
        create_timestamp: float | None = None,
    ) -> "DefsStateInfo":
        new_info = (
            DefsKeyStateInfo(version=version, create_timestamp=create_timestamp or time.time())
            if version
            else None
        )
        if current_info is None:
            return DefsStateInfo(info_mapping={key: new_info})
        else:
            return DefsStateInfo(info_mapping={**current_info.info_mapping, key: new_info})

    def get_version(self, key: str) -> str | None:
        info = self.info_mapping.get(key)
        return info.version if info else None

    def for_keys(self, keys: set[str]) -> "DefsStateInfo":
        """Subsets the DefsStateInfo to only include the keys in the set."""
        return DefsStateInfo(info_mapping={k: v for k, v in self.info_mapping.items() if k in keys})

    @classmethod
    def from_graphql(cls, data: dict[str, Any]) -> "DefsStateInfo":
        return cls(
            info_mapping={
                val["name"]: DefsKeyStateInfo.from_graphql(val["info"])
                for val in data["keyStateInfo"]
            }
        )



import os
import subprocess
import warnings
from pathlib import Path

import dagster_shared
import pytest
from dagster_shared.libraries import (
    DagsterLibraryRegistry,
    check_dagster_package_version,
    core_version_from_library_version,
    library_version_from_core_version,
)

EXCLUDE_LIBRARIES = []


def test_library_registry():
    assert DagsterLibraryRegistry.get() == {
        "dagster-shared": dagster_shared.__version__,
    }


def test_non_dagster_library_registry(library_registry_fixture):
    DagsterLibraryRegistry.register("not-dagster", "0.0.1", is_dagster_package=False)

    assert DagsterLibraryRegistry.get() == {
        "dagster-shared": dagster_shared.__version__,
        "not-dagster": "0.0.1",
    }


def test_all_libraries_register() -> None:
    # attempt to ensure all libraries in the repository register with DagsterLibraryRegistry
    register_call = "DagsterLibraryRegistry.register"

    library_dir = Path(__file__).parents[3] / "libraries"
    assert str(library_dir).endswith("python_modules/libraries")

    for library in os.listdir(library_dir):
        if (
            library in EXCLUDE_LIBRARIES
            or library.startswith(".")
            or library.endswith("CONTRIBUTING.md")
        ):
            continue
        result = subprocess.run(["grep", register_call, (library_dir / library), "-r"], check=False)
        assert result.returncode == 0, (
            f"Dagster library {library} is missing call to {register_call}."
        )


@pytest.fixture
def library_registry_fixture():
    previous_libraries = DagsterLibraryRegistry.get()

    yield

    DagsterLibraryRegistry._libraries = previous_libraries  # noqa: SLF001  # pyright: ignore[reportAttributeAccessIssue]


def test_check_dagster_package_version(monkeypatch):
    monkeypatch.setattr(dagster_shared.version, "__version__", "1.1.0")  # type: ignore

    # Ensure no warning emitted
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        check_dagster_package_version("foo", "1.1.0")

    # Lib version matching 1.1.0-- see dagster._utils.library_version_from_core_version
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        check_dagster_package_version("foo", "0.17.0")

    with pytest.warns(Warning):  # minor version
        check_dagster_package_version("foo", "1.2.0")

    with pytest.warns(Warning):  # patch version
        check_dagster_package_version("foo", "1.1.1")

    with pytest.warns(Warning):  # minor version
        check_dagster_package_version("foo", "0.18.0")

    with pytest.warns(Warning):  # patch version
        check_dagster_package_version("foo", "0.17.1")


def test_library_version_from_core_version():
    assert library_version_from_core_version("1.1.16") == "0.17.16"
    assert library_version_from_core_version("0.17.16") == "0.17.16"
    assert library_version_from_core_version("1.1.16pre0") == "0.17.16rc0"
    assert library_version_from_core_version("1.1.16rc0") == "0.17.16rc0"
    assert library_version_from_core_version("1.1.16post0") == "0.17.16post0"
    assert library_version_from_core_version("1!0+dev") == "1!0+dev"


def test_core_version_from_library_version():
    assert core_version_from_library_version("1!0+dev") == "1!0+dev"
    assert core_version_from_library_version("0.17.16") == "1.1.16"
    assert core_version_from_library_version("1.1.16") == "1.1.16"
    assert core_version_from_library_version("0.17.16pre0") == "1.1.16rc0"
    assert core_version_from_library_version("0.17.16rc0") == "1.1.16rc0"
    assert core_version_from_library_version("0.17.16post0") == "1.1.16post0"



from typing import NamedTuple

import pytest
from dagster_shared.utils.hash import hash_collection


def test_hash_collection():
    # lists have different hashes depending on order
    assert hash_collection([1, 2, 3]) == hash_collection([1, 2, 3])
    assert hash_collection([1, 2, 3]) != hash_collection([2, 1, 3])

    # dicts have same hash regardless of order
    assert hash_collection({"a": 1, "b": 2}) == hash_collection({"b": 2, "a": 1})

    assert hash_collection(set(range(10))) == hash_collection(set(range(10)))

    with pytest.raises(AssertionError):
        hash_collection(object())  # pyright: ignore[reportArgumentType]

    class Foo(NamedTuple):
        a: list[int]
        b: dict[str, int]
        c: str

    with pytest.raises(Exception):
        hash(Foo([1, 2, 3], {"a": 1}, "alpha"))

    class Bar(Foo):
        def __hash__(self):
            return hash_collection(self)

    assert hash(Bar([1, 2, 3], {"a": 1}, "alpha")) == hash(Bar([1, 2, 3], {"a": 1}, "alpha"))



import contextlib
import os
import random
import socket
from collections.abc import Iterator, Mapping
from pathlib import Path
from types import GenericAlias
from typing import TypeGuard, TypeVar

try:
    import tomllib  # pyright: ignore[reportMissingImports]
except ImportError:
    # Python < 3.11 fallback
    import tomli as tomllib

T = TypeVar("T")


def _find_free_port_in_range(start: int, end: int) -> int:
    ports_to_try = list(range(start, end + 1))
    with contextlib.closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        random.shuffle(ports_to_try)
        for port in ports_to_try:
            try:
                s.bind(("", port))
                s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                return port
            except OSError:
                continue

        raise Exception(f"No free ports found in range {start}-{end}")


def find_free_port() -> int:
    port_range = os.getenv("DAGSTER_PORT_RANGE")
    if port_range:
        split_range = port_range.split("-")
        if len(split_range) != 2:
            raise Exception("DAGSTER_PORT_RANGE must be of the form 'start-end'")
        return _find_free_port_in_range(int(split_range[0]), int(split_range[1]))

    with contextlib.closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


def remove_none_recursively(obj: T) -> T:
    """Remove none values from a dict. This can be used to support comparing provided config vs.
    config we retrieve from kubernetes, which returns all fields, even those which have no value
    configured.
    """
    if isinstance(obj, (list, tuple, set)):
        return type(obj)(remove_none_recursively(x) for x in obj if x is not None)
    elif isinstance(obj, dict):
        return type(obj)(
            (remove_none_recursively(k), remove_none_recursively(v))
            for k, v in obj.items()
            if k is not None and v is not None
        )
    else:
        return obj


@contextlib.contextmanager
def environ(env: Mapping[str, str]) -> Iterator[None]:
    """Temporarily set environment variables inside the context manager and
    fully restore previous environment afterwards.
    """
    previous_values = {key: os.getenv(key) for key in env}
    for key, value in env.items():
        if value is None:
            if key in os.environ:
                del os.environ[key]
        else:
            os.environ[key] = value
    try:
        yield
    finally:
        for key, value in previous_values.items():
            if value is None:
                if key in os.environ:
                    del os.environ[key]
            else:
                os.environ[key] = value


def get_boolean_string_value(tag_value: str):
    return tag_value.lower() not in {"false", "none", "0", ""}


def safe_is_subclass(obj, cls: type[T]) -> TypeGuard[type[T]]:
    return (
        isinstance(obj, type)
        and not isinstance(obj, GenericAlias)  # prevent exceptions on 3.9
        and issubclass(obj, cls)
    )


def _read_pyproject_toml(directory: Path | str) -> dict | None:
    """Read and parse pyproject.toml from a directory. Returns None if not found."""
    if isinstance(directory, str):
        directory = Path(directory)
    pyproject_path = directory / "pyproject.toml"
    if not pyproject_path.exists():
        return None
    with open(pyproject_path, "rb") as f:
        return tomllib.load(f)


def find_uv_workspace_root(start_path: Path | str) -> tuple[Path, dict] | None:
    """Walk up directories to find a uv workspace root.

    Returns (workspace_root_path, workspace_config) or None if not found.
    A uv workspace is identified by [tool.uv.workspace] in pyproject.toml.
    """
    if isinstance(start_path, str):
        start_path = Path(start_path)
    current = start_path.resolve()
    while True:
        data = _read_pyproject_toml(current)
        if data:
            workspace_config = data.get("tool", {}).get("uv", {}).get("workspace", {})
            if workspace_config:
                return current, workspace_config
        parent = current.parent
        if parent == current:
            return None
        current = parent



import os
import posixpath
from pathlib import Path
from typing import Any

DEFAULT_FILE_EXCLUDE_PATTERNS: list[str] = [
    "__pycache__",
    ".pytest_cache",
    "*.egg-info",
    "*.cpython-*",
    ".DS_Store",
    ".ruff_cache",
    "tox.ini",
    ".gitkeep",  # dummy file that allows empty directories to be checked into git
]

PROJECT_NAME_PLACEHOLDER = "PROJECT_NAME_PLACEHOLDER"


def should_skip_scaffolded_file(path: str, excludes: list[str] = DEFAULT_FILE_EXCLUDE_PATTERNS):
    """Given a file path `path` in a source template, returns whether or not the file should be skipped
    when generating destination files.

    Technically, `path` could also be a directory path that should be skipped.
    """
    for pattern in excludes:
        if pattern.lower() in path.lower():
            return True

    return False


def scaffold_subtree(
    *,
    path: Path,
    project_template_path: Path,
    excludes: list[str] | None = None,
    name_placeholder: str = PROJECT_NAME_PLACEHOLDER,
    project_name: str | None = None,
    **other_template_vars: Any,
):
    """Renders templates for Dagster project."""
    # defer for import performance
    import jinja2

    excludes = (
        DEFAULT_FILE_EXCLUDE_PATTERNS if not excludes else DEFAULT_FILE_EXCLUDE_PATTERNS + excludes
    )

    normalized_path = os.path.normpath(path)
    project_name = project_name or os.path.basename(normalized_path).replace("-", "_")
    if not os.path.exists(normalized_path):
        os.makedirs(normalized_path, exist_ok=True)

    loader = jinja2.FileSystemLoader(searchpath=project_template_path)
    env = jinja2.Environment(loader=loader)

    # merge custom skip_files with the default list
    for root, dirs, files in os.walk(project_template_path):
        # For each subdirectory in the source template, create a subdirectory in the destination.
        for dirname in dirs:
            src_dir_path = os.path.join(root, dirname)
            if should_skip_scaffolded_file(src_dir_path, excludes):
                continue

            src_relative_dir_path = os.path.relpath(src_dir_path, project_template_path)
            dst_relative_dir_path = src_relative_dir_path.replace(
                name_placeholder,
                project_name,
                1,
            )
            dst_dir_path = os.path.join(normalized_path, dst_relative_dir_path)

            os.mkdir(dst_dir_path)

        # For each file in the source template, render a file in the destination.
        for filename in files:
            src_file_path = os.path.join(root, filename)
            if should_skip_scaffolded_file(src_file_path, excludes):
                continue

            src_relative_file_path = os.path.relpath(src_file_path, project_template_path)
            dst_relative_file_path = src_relative_file_path.replace(
                name_placeholder,
                project_name,
                1,
            )
            dst_file_path = os.path.join(normalized_path, dst_relative_file_path)

            if dst_file_path.endswith(".jinja"):
                dst_file_path = dst_file_path[: -len(".jinja")]

            with open(dst_file_path, "w", encoding="utf8") as f:
                # Jinja template names must use the POSIX path separator "/".
                template_name = src_relative_file_path.replace(os.sep, posixpath.sep)
                template: jinja2.environment.Template = env.get_template(name=template_name)
                f.write(
                    template.render(
                        repo_name=project_name,  # deprecated
                        project_name=project_name,
                        **other_template_vars,
                    )
                )
                f.write("\n")



import inspect
import os
import sys
from abc import ABC
from collections import namedtuple
from collections.abc import Callable, Iterator, Mapping
from functools import partial
from typing import TYPE_CHECKING, Any, NamedTuple, TypeVar, overload

from typing_extensions import Self, dataclass_transform

import dagster_shared.check as check
from dagster_shared.check.builder import (
    INJECTED_CHECK_VAR,
    INJECTED_DEFAULT_VALS_LOCAL_VAR,
    EvalContext,
    build_args_and_assignment_strs,
    build_check_call_str,
)
from dagster_shared.check.record import RECORD_MARKER_FIELD, RECORD_MARKER_VALUE, is_record

ImportFrom = check.ImportFrom  # re-expose for convenience
TType = TypeVar("TType", bound=type)
TVal = TypeVar("TVal")


_RECORD_ANNOTATIONS_FIELD = "__record_annotations__"
_RECORD_DEFAULTS_FIELD = "__record_defaults__"
_CHECKED_NEW = "__checked_new__"
_DEFAULTS_NEW = "__defaults_new__"
_NAMED_TUPLE_BASE_NEW_FIELD = "__nt_new__"
_REMAPPING_FIELD = "__field_remap__"
_ORIGINAL_CLASS_FIELD = "__original_class__"
_KW_ONLY_FIELD = "__kw_only__"
_REPLACE_NEW_FIELD = "__replace_new__"

_sample_nt = namedtuple("_canary", "x")
# use a sample to avoid direct private imports (_collections._tuplegetter)
_tuple_getter_type = type(getattr(_sample_nt, "x"))


def _get_field_set_and_defaults(
    cls: type,
    kw_only: bool,
) -> tuple[Mapping[str, Any], Mapping[str, Any]]:
    field_set = getattr(cls, "__annotations__", {})
    defaults = {}

    last_defaulted_field = None
    for name in field_set.keys():
        if hasattr(cls, name):
            attr_val = getattr(cls, name)
            if isinstance(attr_val, property):
                check.invariant(
                    attr_val.__isabstractmethod__,
                    f"Conflicting non-abstract @property for field {name} on record {cls.__name__}."
                    "Add the the @abstractmethod decorator to make it abstract.",
                )
            # When doing record inheritance, filter out tuplegetters from parents.
            # This workaround only seems needed for py3.9
            elif not isinstance(attr_val, _tuple_getter_type):
                check.invariant(
                    not inspect.isfunction(attr_val),
                    f"Conflicting function for field {name} on record {cls.__name__}. "
                    "If you are trying to set a function as a default value "
                    "you will have to override __new__.",
                )
                if "pydantic" in sys.modules:
                    from pydantic.fields import FieldInfo

                    check.invariant(
                        not isinstance(attr_val, FieldInfo),
                        "pydantic.Field is not supported as a default value for @record fields."
                        " For Resolved subclasses, you may provide additional field metadata"
                        " through the Resolver: Annotated[..., Resolver.default(description=...)].",
                    )
                defaults[name] = attr_val
                last_defaulted_field = name
                continue

        # fall through here means no default set
        if last_defaulted_field and not kw_only:
            check.failed(
                "Fields without defaults cannot appear after fields with default values. "
                f"Field {name} has no default after {last_defaulted_field} with default value."
            )

    for base in cls.__bases__:
        if is_record(base):
            original_base = getattr(base, _ORIGINAL_CLASS_FIELD)
            base_kw_only = getattr(base, _KW_ONLY_FIELD)
            check.invariant(
                kw_only == base_kw_only,
                "Can not inherit from a parent @record with different kw_only setting.",
            )
            base_field_set, base_defaults = _get_field_set_and_defaults(original_base, kw_only)
            field_set = {**base_field_set, **field_set}
            defaults = {**base_defaults, **defaults}

    return field_set, defaults


def _namedtuple_record_transform(
    cls: TType,
    *,
    checked: bool,
    with_new: bool,
    decorator_frames: int,
    field_to_new_mapping: Mapping[str, str] | None,
    kw_only: bool,
) -> TType:
    """Transforms the input class in to one that inherits a generated NamedTuple base class
    and:
        * bans tuple methods that don't make sense for a record object
        * creates a run time checked __new__  (optional).
    """
    field_set, defaults = _get_field_set_and_defaults(cls, kw_only)

    base = NamedTuple(f"_{cls.__name__}", field_set.items())
    nt_new = base.__new__

    generated_new = None
    if checked:
        eval_ctx = EvalContext.capture_from_frame(
            1 + decorator_frames,
            add_to_local_ns={
                # inject default values in to the local namespace for reference in generated __new__
                INJECTED_DEFAULT_VALS_LOCAL_VAR: defaults,
                # as well as a ref to the check module
                INJECTED_CHECK_VAR: check,
            },
        )
        generated_new = JitCheckedNew(
            field_set,
            defaults,
            eval_ctx,
            new_frames=1 if with_new else 0,
            kw_only=kw_only,
        )
    elif defaults:
        # allow arbitrary ordering of default values by generating a kwarg only __new__ impl
        eval_ctx = EvalContext(
            global_ns={},
            # inject default values in to the local namespace for reference in generated __new__
            local_ns={INJECTED_DEFAULT_VALS_LOCAL_VAR: defaults},
            lazy_imports={},
        )
        generated_new = eval_ctx.compile_fn(
            _build_defaults_new(field_set, defaults, kw_only),
            _DEFAULTS_NEW,
        )

    # the default namedtuple record cannot handle subclasses that have different fields from their
    # parents if both are records
    base.__repr__ = _repr
    nt_iter = base.__iter__

    # disable iteration unless the debugger is running which uses iteration to display values
    if not os.getenv("DEBUGPY_RUNNING"):
        base.__iter__ = _banned_iter
        base.__getitem__ = _banned_idx

    # these will override an implementation on the class if it exists
    new_class_dict = {
        **{n: getattr(base, n) for n in field_set.keys()},
        "_fields": base._fields,
        "__hidden_iter__": nt_iter,
        "__hidden_replace__": base._replace,
        RECORD_MARKER_FIELD: RECORD_MARKER_VALUE,
        _RECORD_ANNOTATIONS_FIELD: field_set,
        _RECORD_DEFAULTS_FIELD: defaults,
        _NAMED_TUPLE_BASE_NEW_FIELD: nt_new,
        _REMAPPING_FIELD: field_to_new_mapping or {},
        _ORIGINAL_CLASS_FIELD: cls,
        _KW_ONLY_FIELD: kw_only,
        _REPLACE_NEW_FIELD: generated_new if generated_new else nt_new,
        "__reduce__": _reduce,
        # functools doesn't work, so manually update_wrapper
        "__module__": cls.__module__,
        "__qualname__": cls.__qualname__,
        "__annotations__": field_set,
        "__doc__": cls.__doc__,
        "__get_pydantic_core_schema__": _pydantic_core_schema,
    }

    # Due to MRO issues, we can not support both @record_custom __new__ and record inheritance
    # so enforce these rules and place the generated new in the appropriate place in the class hierarchy.
    if with_new:
        if not _defines_own_new(cls):
            # verify the alignment since it impacts frame capture
            check.failed(
                f"Expected __new__ on {cls.__name__}, add it or switch from the @record_custom decorator to @record."
            )
        if is_record(cls):
            parent = next(c for c in cls.__mro__ if is_record(c))
            check.failed(
                f"@record_custom can not be used with @record inheritance. {cls.__name__} is a child of @record {parent.__name__}."
            )

        # For records with custom new, put the generated new on the NT base class
        if generated_new:
            base.__new__ = generated_new

    elif generated_new:
        if _defines_own_new(cls):
            check.failed(
                f"Found a custom __new__ on @record {cls}, use @record_custom with IHaveNew instead."
            )
        for c in cls.__mro__:
            if c is cls:
                continue

            if is_record(c) and _defines_own_new(c):
                check.failed(
                    "@record can not inherit from @record_custom. "
                    f"@record {cls.__name__} inherits from @record_custom {c.__name__}"
                )

        # For regular @records, which may inherit from other records, put new on the generated class to avoid
        # MRO resolving to the wrong NT base
        new_class_dict["__new__"] = generated_new
        new_class_dict["_make"] = base._make

    # make it so instances of records with no fields still evaluate to true
    if len(field_set) < 1:
        new_class_dict["__bool__"] = _true

    new_type = type(
        cls.__name__,
        (cls, base),
        new_class_dict,
    )

    # setting this in the dict above does not work for some reason,
    # so set it directly after instantiation
    if hasattr(cls, "__parameters__"):
        setattr(new_type, "__parameters__", cls.__parameters__)

    return new_type  # type: ignore


@overload
def record(
    cls: TType,
) -> TType: ...  # Overload for using decorator with no ().


@overload
def record(
    *,
    checked: bool = True,
    kw_only: bool = True,
) -> Callable[[TType], TType]: ...  # Overload for using decorator used with args.


@dataclass_transform(
    kw_only_default=True,
    frozen_default=True,
)
def record(
    cls: TType | None = None,
    *,
    checked: bool = True,
    kw_only: bool = True,
) -> TType | Callable[[TType], TType]:
    """A class decorator that will create an immutable record class based on the defined fields.

    Args:
        checked: Whether or not to generate runtime type checked construction (default True).
        kw_only: Whether or not the generated __new__ is kwargs only (default True).

    Example:
        @record
        class Person:
            name: str
            age: int

        # Create instance
        person = Person(name="Alice", age=30)

        # Create modified copy using replace()
        older_person = replace(person, age=31)

        # Add items to lists using replace()
        @record
        class TaskList:
            items: list[str]

        tasks = TaskList(items=["task1"])
        updated_tasks = replace(tasks, items=[*tasks.items, "task2"])
    """
    if cls:
        return _namedtuple_record_transform(
            cls,
            checked=checked,
            with_new=False,
            decorator_frames=1,
            field_to_new_mapping=None,
            kw_only=kw_only,
        )
    else:
        return partial(
            _namedtuple_record_transform,
            checked=checked,
            with_new=False,
            decorator_frames=0,
            field_to_new_mapping=None,
            kw_only=kw_only,
        )


@overload
def record_custom(
    cls: TType,
) -> TType: ...  # Overload for using decorator with no ().


@overload
def record_custom(
    *,
    checked: bool = True,
    field_to_new_mapping: Mapping[str, str] | None = None,
) -> Callable[[TType], TType]: ...  # Overload for using decorator used with args.


def record_custom(
    cls: TType | None = None,
    *,
    checked: bool = True,
    field_to_new_mapping: Mapping[str, str] | None = None,
) -> TType | Callable[[TType], TType]:
    """Variant of the record decorator to use when overriding __new__.

    Example:
        @record_custom
        class MyRecord(IHaveNew):
            name: str
            value: int

            def __new__(cls, name: str, value: int = 42):
                # Custom logic here
                if not name:
                    name = "default"
                return super().__new__(
                    cls,
                    name=name,
                    value=value,
                )

    Important requirements:
    - Must inherit from IHaveNew
    - Must define a custom __new__ method
    - Must call super().__new__(cls, **kwargs) with keyword arguments
    - Keyword arguments must match the field names exactly
    - Cannot be used with @record inheritance

    This approach is useful when you need custom validation, default value logic,
    or type coercion in the constructor that can't be handled with simple defaults.

    It would have been cool if we could do that with an argument and @overload but
    from https://peps.python.org/pep-0681/ "When applied to an overload,
    the dataclass_transform decorator still impacts all usage of the function."
    """
    if cls:
        return _namedtuple_record_transform(
            cls,
            checked=checked,
            with_new=True,
            decorator_frames=1,
            field_to_new_mapping=field_to_new_mapping,
            kw_only=True,
        )
    else:
        return partial(
            _namedtuple_record_transform,
            checked=checked,
            with_new=True,
            decorator_frames=0,
            field_to_new_mapping=field_to_new_mapping,
            kw_only=True,
        )


class IHaveNew:
    """Marker class to be used when overriding new in @record_custom classes to prevent
    type errors when calling super().__new__.
    """

    if TYPE_CHECKING:

        def __new__(cls, **kwargs) -> Self: ...

        # let type checker know these objects are sortable (by way of being a namedtuple)
        def __lt__(self, other) -> bool: ...


def has_generated_new(obj) -> bool:
    return obj.__new__.__name__ in (_DEFAULTS_NEW, _CHECKED_NEW)


def get_record_annotations(obj) -> Mapping[str, type]:
    check.invariant(is_record(obj), "Only works for @record decorated classes")
    return getattr(obj, _RECORD_ANNOTATIONS_FIELD)


def get_record_defaults(obj) -> Mapping[str, Any]:
    check.invariant(is_record(obj), "Only works for @record decorated classes")
    return getattr(obj, _RECORD_DEFAULTS_FIELD)


def get_original_class(obj):
    check.invariant(is_record(obj), "Only works for @record decorated classes")
    return getattr(obj, _ORIGINAL_CLASS_FIELD)


def as_dict(obj) -> Mapping[str, Any]:
    """Creates a dict representation of the record based on the fields."""
    check.invariant(is_record(obj), "Only works for @record decorated classes")

    return {key: value for key, value in zip(obj._fields, obj.__hidden_iter__())}


def as_dict_for_new(obj) -> Mapping[str, Any]:
    """Creates a dict representation of the record with field_to_new_mapping applied."""
    check.invariant(is_record(obj), "Only works for @record decorated classes")

    remap = getattr(obj, _REMAPPING_FIELD)
    from_obj = {}
    for k, v in as_dict(obj).items():
        if k in remap:
            from_obj[remap[k]] = v
        else:
            from_obj[k] = v

    return from_obj


def copy(record: TVal, **kwargs) -> TVal:
    """Create a new instance of this record type using its constructor with
    the original records values overrode with new values specified as key args.
    """
    return record.__class__(
        **{
            **as_dict_for_new(record),
            **kwargs,
        }
    )


def replace(obj: TVal, **kwargs) -> TVal:
    """Create a new instance of this record type using the record constructor directly,
    (bypassing any custom __new__ impl) with the original records values overrode with
    new values specified by keyword args.

    This emulates the behavior of namedtuple _replace.

    Example:
        @record
        class Config:
            timeout: int
            retries: int

        config = Config(timeout=30, retries=3)

        # Update single field
        faster_config = replace(config, timeout=10)

        # Update multiple fields
        production_config = replace(config, timeout=60, retries=5)

        # Add items to lists (use spread operator to avoid mutation)
        @record
        class Results:
            errors: list[str]
            warnings: list[str]

        results = Results(errors=[], warnings=["deprecated API"])

        # Add new error
        updated_results = replace(results, errors=[*results.errors, "network timeout"])

        # Add multiple items
        final_results = replace(
            updated_results,
            errors=[*updated_results.errors, "connection failed"],
            warnings=[*updated_results.warnings, "slow query"]
        )
    """
    check.invariant(is_record(obj))
    cls = obj.__class__

    # if we have runtime type checking, go through that to vet new field values
    return getattr(cls, _REPLACE_NEW_FIELD)(
        cls,
        **{**as_dict(obj), **kwargs},
    )


class LegacyNamedTupleMixin(ABC):
    """Mixin to ease migration by adding NamedTuple utility methods.
    Inherit when converting an existing NamedTuple that has callsites to _replace / _asdict, ie.

    @record
    def AssetSubset(LegacyNamedTupleMixin):
        asset_key: AssetKey
        value: Union[bool, PartitionsSubset]
    """

    def _replace(self, **kwargs) -> Self:
        return replace(self, **kwargs)

    def _asdict(self) -> Mapping[str, Any]:
        return as_dict(self)

    def __iter__(self) -> Iterator:
        return tuple.__iter__(self)  # type: ignore


class JitCheckedNew:
    """Object that allows us to just-in-time compile a checked __new__ implementation on first use.
    This has two benefits:
        1. Defer processing ForwardRefs until their definitions are in scope.
        2. Avoid up-front cost for unused objects.
    """

    __name__ = _CHECKED_NEW

    def __init__(
        self,
        field_set: Mapping[str, type],
        defaults: Mapping[str, Any],
        eval_ctx: EvalContext,
        new_frames: int,
        kw_only: bool,
    ):
        self._field_set = field_set
        self._defaults = defaults
        self._eval_ctx = eval_ctx
        self._new_frames = new_frames  # how many frames of __new__ there are
        self._compiled = False
        self._kw_only = kw_only

    def __call__(self, cls, *args, **kwargs):
        if _do_defensive_checks():
            # this condition can happen during races in threaded envs so only
            # invariant when opted-in
            check.invariant(
                self._compiled is False,
                "failed to set compiled __new__ appropriately",
            )

        # update the context with callsite locals/globals to resolve
        # ForwardRefs that were unavailable at definition time.
        self._eval_ctx.update_from_frame(1 + self._new_frames)

        # we are double-memoizing this to handle some confusing mro issues
        # in which the _nt_base's __new__ method is not on the critical
        # path, causing this to get invoked multiple times
        compiled_fn = self._eval_ctx.compile_fn(
            self._build_checked_new_str(),
            _CHECKED_NEW,
        )
        self._compiled = True

        # replace this holder object with the compiled fn by finding where it was in the hierarchy
        for c in cls.__mro__:
            if c.__new__ is self:
                c.__new__ = compiled_fn
        setattr(cls, _REPLACE_NEW_FIELD, compiled_fn)

        return compiled_fn(cls, *args, **kwargs)

    def _build_checked_new_str(self) -> str:
        args_str, set_calls_str = build_args_and_assignment_strs(
            self._field_set.keys(),
            self._defaults,
            self._kw_only,
        )
        check_calls = []
        for name, ttype in self._field_set.items():
            call_str = build_check_call_str(
                ttype=ttype,
                name=name,
                eval_ctx=self._eval_ctx,
            )
            check_calls.append(f"{name}={call_str}")

        check_call_block = ",\n        ".join(check_calls)

        lazy_imports_str = "\n    ".join(
            f"from {module} import {t}" for t, module in self._eval_ctx.lazy_imports.items()
        )

        checked_new_str = f"""
def __checked_new__(cls{args_str}):
    {lazy_imports_str}
    {set_calls_str}
    return cls.{_NAMED_TUPLE_BASE_NEW_FIELD}(
        cls,
        {check_call_block}
    )
"""
        return checked_new_str


def _build_defaults_new(
    field_set: Mapping[str, type],
    defaults: Mapping[str, Any],
    kw_only: bool,
) -> str:
    """Build a __new__ implementation that handles default values."""
    kw_args_str, set_calls_str = build_args_and_assignment_strs(
        field_set,
        defaults,
        kw_only,
    )
    assign_str = ",\n        ".join([f"{name}={name}" for name in field_set.keys()])
    return f"""
def __defaults_new__(cls{kw_args_str}):
    {set_calls_str}
    return cls.{_NAMED_TUPLE_BASE_NEW_FIELD}(
        cls,
        {assign_str}
    )
    """


def _banned_iter(self, *args, **kwargs):
    raise Exception(f"Iteration is not allowed on `@record` {self.__class__.__name__}.")


def _banned_idx(self, *args, **kwargs):
    raise Exception(f"Index access is not allowed on `@record` {self.__class__.__name__}.")


def _true(_):
    return True


def _from_reduce(cls, kwargs):
    # loading from pickle bypasses checked / custom __new__ and
    # just reconstructs the base namedtuple
    return getattr(cls, _NAMED_TUPLE_BASE_NEW_FIELD)(
        cls,
        **kwargs,
    )


def _reduce(self):
    # pickle support
    return _from_reduce, (self.__class__, as_dict(self))


def _repr(self) -> str:
    # the __repr__ method generated for namedtuples cannot handle subclasses that have different
    # sets of fields
    field_set = getattr(self, _RECORD_ANNOTATIONS_FIELD)
    values = [f"{field_name}={getattr(self, field_name)!r}" for field_name in field_set]
    return f"{self.__class__.__name__}({', '.join(values)})"


def _defines_own_new(cls) -> bool:
    qualname = getattr(cls.__new__, "__qualname__", None)
    if not qualname:
        return False

    qualname_parts = cls.__new__.__qualname__.split(".")
    if len(qualname_parts) < 2:
        return False

    return qualname_parts[-2] == cls.__name__


def _do_defensive_checks():
    return bool(os.getenv("DAGSTER_RECORD_DEFENSIVE_CHECKS"))


@classmethod
def _pydantic_core_schema(cls, source: Any, handler):
    """Forces pydantic_core to treat records as normal types instead of namedtuples. In particular,
    pydantic assumes that all namedtuples can be constructed with posargs, while records are kw-only.
    """
    from pydantic_core import core_schema

    return core_schema.is_instance_schema(cls)
